@@ -6,13 +6,15 @@
  ************************************************************************/
 #ifdef RCS
 static /*const*/char rcsid[]=
- "$Id: atotime.c,v 1.3 1997/08/01 01:57:01 srb Exp $";
+ "$Id: atotime.c,v 1.5 1998/05/02 00:56:54 srb Exp $";
 #endif
 #include <unistd.h>
 #include <string.h>
 #include <time.h>
 #include <stdlib.h>
 
+#include "atotime.h"
+
 static unsigned dec(const char*const s)
 { return strtol(s,(char**)0,10);
 }
@@ -37,7 +39,7 @@ time_t atosec(const char*in,const char**const innew)
      static time_t tscale[]={1,60,60*60,60*60*24,60*60*24*7,
       60*60*24*(365*4+1)/4/12,60*60*24*(365*4+1)/4};
      time_t i=strtol(p=in,(char**)&in,10);
-     if(p!=in&&(p=strchr(scale,*in)))
+     if(p!=in&&*in&&(p=strchr(scale,*in)))
 	in++,t+=i*tscale[(p-scale)/2];
      else
       { *innew=in;
