@@ -11,7 +11,7 @@
  ************************************************************************/
 #ifdef RCS
 static /*const*/char rcsid[]=
- "$Id: authenticate.c,v 1.10 1997/08/01 01:00:31 srb Exp $";
+ "$Id: authenticate.c,v 1.19 1998/05/13 16:57:39 srb Exp $";
 #endif
 
 #ifdef PROCMAIL
@@ -22,6 +22,8 @@ static /*const*/char rcsid[]=
 #else
 #include "config.h"
 
+#define _XOPEN_SOURCE
+
 #include <sys/types.h>
 #include <unistd.h>
 #include <pwd.h>
@@ -35,6 +37,12 @@ static /*const*/char rcsid[]=
 
 #include "authenticate.h"
 
+#define VIRTUALSERVER	/**/
+
+#define SENDMAILLIB	"/etc/mail"
+#define VIRTUALHOSTDB	"vpop.db"
+#define VIRTUALUSER	"vpop"
+
 #ifndef MAILSPOOLDIR
 #define MAILSPOOLDIR	"/var/spool/mail/"	     /* watch the trailing / */
 #endif
@@ -43,14 +51,27 @@ static /*const*/char rcsid[]=
 #endif
 /*#define MAILSPOOLHOME "/.mail"		      /* watch the leading / */
 						  /* delivers to $HOME/.mail */
+#ifdef VIRTUALSERVER
+#include <netdb.h>
+#include <netinet/in.h>
+
+#include "sdb.h"
+#include "simplecrypt.h"
+
+const char virtualhostdb[]=SENDMAILLIB"/"VIRTUALHOSTDB,
+ virtualuser[]=VIRTUALUSER;
+#endif /* VIRTUALSERVER */
+
 #define STRLEN(x)	(sizeof(x)-1)
 
 struct auth_identity
 { const struct passwd*pw;
-  char*mbox;
+  char*mbox,*usersecret;
   int sock;
 };
 
+char*auth_logname;
+
 static auth_identity authi;		      /* reuse copy, only one active */
 
 static void castlower(str)register char*str;   /* and I'll take the low road */
@@ -59,17 +80,84 @@ static void castlower(str)register char*str;   /* and I'll take the low road */
 	*str+='a'-'A';				     /* cast it to lowercase */
 }
 
-static const struct passwd*cgetpwnam(user,sock)const char*const user;
+static void getlogname(user,sock)const char*user;const int sock;
+{ struct sockaddr_in sockname;int namelen=sizeof sockname;const char*retval="";
+  if(!getsockname(sock,(struct sockaddr*)&sockname,&namelen))
+   { const struct hostent*hent=
+      gethostbyaddr((void*)&sockname.sin_addr.s_addr,sizeof sockname.sin_addr,
+       sockname.sin_family);
+     if(hent)
+	retval=hent->h_name;
+   }
+  if(auth_logname=			   /* memory leak when out of memory */
+   realloc(auth_logname,(namelen=strlen(retval))+1+strlen(user)+1))
+   { strcpy(auth_logname,retval)[namelen]='/';
+     strcpy(auth_logname+namelen+1,user);
+   }
+}
+
+static const struct passwd*cgetpwnam(user,sock)const char*user;
  const int sock;
-{ return getpwnam(user);	       /* this should be selfexplanatory :-) */
+{
+#ifdef VIRTUALSERVER
+  DB_ENV dbenv;DB*db;
+  memset(&dbenv,0,sizeof dbenv);
+  if(!db_appinit(0,0,&dbenv,0))
+   { if(!db_open(virtualhostdb,DB_HASH,DB_RDONLY,0666,&dbenv,(void*)0,&db))
+      { getlogname(user,sock);
+	if(auth_logname)
+	 { DBT k,d;unsigned foundentry=1;
+	   memset(&k,0,sizeof k);memset(&d,0,sizeof d);
+	   k.size=strlen(k.data=auth_logname);
+	   if(db->get(db,0,&k,&d,0))
+	    { k.size=strlen(strcpy(auth_logname,user));
+	      foundentry=!db->get(db,0,&k,&d,0)<<1;	   /* fallback check */
+	    }
+	   if(foundentry&&(authi.usersecret=malloc(d.size+1)))
+	    { scmorph(&k,&d);memcpy(authi.usersecret,d.data,d.size);
+	      authi.usersecret[d.size]='\0';memset(d.data,0,d.size);
+	      if(foundentry&2)
+		 goto freesec;
+	      else if(*authi.usersecret)	 /* no empty secrets allowed */
+		 user=virtualuser;
+	      else
+freesec:       { free(authi.usersecret);authi.usersecret=0;
+		 goto novirt;
+	       }
+	    }
+	   else
+novirt:	      strcpy(auth_logname,user);
+	 }
+	db->close(db,0);
+      }					   /* memory leak when out of memory */
+     else if(auth_logname=realloc(auth_logname,strlen(user)+1))
+	strcpy(auth_logname,user);			 /* DB doesn't exist */
+     db_appexit(&dbenv);
+   }
+  else						     /* DB subsystem problem */
+   { if(auth_logname=realloc(auth_logname,strlen(user)+1))
+	strcpy(auth_logname,user);
+   }
+#endif
+  return getpwnam(user);	       /* this should be selfexplanatory :-) */
 }
 
 static const struct passwd*cgetpwuid(uid,sock)const uid_t uid;const int sock;
-{ return getpwuid(uid);					       /* no comment */
+{ const struct passwd*pass;
+  pass=getpwuid(uid);
+#ifdef VIRTUALSERVER
+  ;{ const char*logname=pass->pw_name;
+     if(pass&&!auth_logname&&(auth_logname=malloc(strlen(logname))))
+	strcpy(auth_logname,logname);
+   }
+#endif
+  return pass;
 }
 
 /*const*/auth_identity*auth_finduser(user,sock)char*const user;const int sock;
-{ if(!(authi.pw=cgetpwnam(user,sock)))		  /* /etc/passwd user lookup */
+{ if(authi.usersecret)
+     free(authi.usersecret),authi.usersecret=0;
+  if(!(authi.pw=cgetpwnam(user,sock)))		  /* /etc/passwd user lookup */
    { char*p;
      if(p=strchr(user,'@'))		  /* does the username contain an @? */
 	*p='\0';		      /* clueless user using the mailaddress */
@@ -87,6 +175,8 @@ static const struct passwd*cgetpwuid(uid,sock)const uid_t uid;const int sock;
 { if(!(authi.pw=cgetpwuid(uid,sock)))		  /* /etc/passwd user lookup */
      return 0;							     /* nada */
   authi.sock=sock;		    /* save filedescriptor for later perusal */
+  if(authi.usersecret)
+     free(authi.usersecret),authi.usersecret=0;
   if(authi.mbox)				   /* old mailbox reference? */
      free(authi.mbox),authi.mbox=0;		/* nix old mailbox reference */
   return &authi;					       /* user found */
@@ -109,7 +199,11 @@ int auth_checkpassword(pass,pw,allowemptypw)const auth_identity*const pass;
 }
 
 const char*auth_getsecret(pass)const auth_identity*const pass;
-{ return 0;	       /* no standard way to get a secret, add function here */
+{ return authi.usersecret;
+}
+
+int auth_checkvalidshell(pass)const auth_identity*const pass;
+{
 }
 #else /* PROCMAIL */
 auth_identity*auth_newid P((void))
@@ -156,19 +250,25 @@ const char*auth_mailboxname(pass)auth_identity*const pass;
      strcpy(pass->mbox,pass->pw->pw_dir);
      strcpy(pass->mbox+i,mailfile);
 #else
-   { static const char mailspooldir[]=MAILSPOOLDIR;
-     if(!(pass->mbox=malloc(
-      STRLEN(mailspooldir)+MAILSPOOLHASH*2+strlen(pass->pw->pw_name)+1)))
+   { static const char mailspooldir[]=MAILSPOOLDIR;size_t servernlen;
+#ifdef VIRTUALSERVER
+     char*logname=auth_logname;
+#else
+     char*logname=pass->pw->pw_name;
+#endif
+     if(!logname||!(pass->mbox=malloc(STRLEN(mailspooldir)+
+      MAILSPOOLHASH*2+strlen(logname)+1)))
 	return "";
      strcpy(pass->mbox,mailspooldir);
      ;{ char*p,*n;size_t i;int c;
-	for(p=pass->mbox+STRLEN(mailspooldir),n=pass->pw->pw_name,
-	 i=MAILSPOOLHASH;i--;*p++='/')
+	for(p=pass->mbox+STRLEN(mailspooldir),n=logname,i=MAILSPOOLHASH;
+	    i--;
+	    *p++='/')
 	 { if(*n)
 	      c= *n++;
 	   *p++=c;
 	 }
-	strcpy(p,pass->pw->pw_name);
+	strcpy(p,logname);
       }
 #endif /* MAILSPOOLHOME */
    }
@@ -192,7 +292,7 @@ const char*auth_shell(pass)const auth_identity*const pass;
 }
 
 const char*auth_username(pass)const auth_identity*const pass;
-{ return pass->pw->pw_name;
+{ return auth_logname?auth_logname:pass->pw->pw_name;
 }
 
 void auth_end P((void))
