@@ -1,4 +1,4 @@
-/*$Id: authenticate.h,v 1.5 1997/04/11 14:08:31 srb Exp $*/
+/*$Id: authenticate.h,v 1.8 1998/05/13 16:57:39 srb Exp $*/
 
 /* Generic authentication interface, substitute a suitable module to
    accomodate arbitrary other authentication databases */
@@ -10,6 +10,8 @@ typedef struct auth_identity auth_identity;
 #define Q(x)	()
 #endif
 
+extern char*auth_logname;
+extern const char virtualhostdb[],virtualuser[];
 /*const*/auth_identity
  *auth_finduser P((char*const user,const sock)),
  *auth_finduid Q((const uid_t uid,const sock));
