@@ -1,4 +1,4 @@
-/*$Id: config.h,v 1.9 1997/08/01 01:13:05 srb Exp $*/
+/*$Id: config.h,v 1.20 1998/05/12 21:09:14 srb Exp $*/
 
 #define USEdot_lock	/**/
 /*#define USEfcntl_lock /**/	    /* to test which combinations make sense */
@@ -12,7 +12,7 @@
 #define APOP		/**/
 #define TOP		/**/
 
-/*#define LAST_HACK	/**/		   /* uncomment to enable dummy LAST */
+#define LAST_HACK	/**/		   /* uncomment to enable dummy LAST */
 							 /* violates RFC1939 */
 
 /* to change the MAILSPOOLDIR, edit authenticate.c */
@@ -69,10 +69,21 @@ typedef unsigned long uidl_t;			    /* for the UIDL wordsize */
 #define LOCKRACE	32
 #define SPOOFMASK	0xffff
 #define AGETOLERANCE	8192	       /* to allow for incorrect system time */
+#define BSIZECATEGORY	16384
+#define DIRECTFETCHAGE	32768
 
 #define CONTLENGTH	"content-length:"
+#define STATUSFIELD	"status:"
+#define XSTATUSFIELD	"x-status:"
 #define ARGLEN		40			 /* as specified by RFC 1939 */
 #define UIDLLEN		70					/* same here */
 #define LINEBUFLEN	(4+1+(ARGLEN+1)*2)    /* maximum command line length */
 #define LISTEN_BACKLOG	128
 #define TCP_PROT	6			      /* from /etc/protocols */
+
+#define MAXBULLETINS	64
+#define MAXSTATEAGE	8388608				       /* > 3 months */
+#define MEMORY_CACHE	(64*1024)
+#define CUCIPOP_LIB	"/var/lib/cucipop"
+#define STATE_DB	"state.db"
+#define BULLETINS_PATH	CUCIPOP_LIB"/bulletins"
