@@ -6,12 +6,12 @@
  *	in-situ mailbox rewrite, no tempfiles, short locking,		*
  *	standard BSD mailbox format, SysV Content-Length extension	*
  *									*
- *	Copyright (c) 1996-1997, S.R. van den Berg, The Netherlands	*
+ *	Copyright (c) 1996-1998, S.R. van den Berg, The Netherlands	*
  *	#include "README"						*
  ************************************************************************/
 #ifdef RCS
 static /*const*/char rcsid[]=
- "$Id: cucipop.c,v 1.13 1997/08/10 10:08:57 srb Exp $";
+ "$Id: cucipop.c,v 1.31 1998/05/13 16:57:39 srb Exp $";
 #endif
 
 #include "patchlevel.h"
@@ -41,6 +41,8 @@ static /*const*/char rcsid[]=
 #include "atotime.h"
 #include "authenticate.h"
 #include "locking.h"
+#include "dbops.h"
+#include "hsort.h"
 
 #ifdef APOP
 #include "md5/md5.h"
@@ -49,15 +51,16 @@ static /*const*/char rcsid[]=
 #define MD5_DIGLEN	16
 #define BIN2HEXLEN	(MD5_DIGLEN*2)
 #define Signal(sig,fn)	signal(sig,(void(*)())(fn))
+#define Vuser		(auth_logname?auth_logname:user)
 
-#define Bintohex(arg)	(sizeof(arg)>MD5_DIGLEN?\
+#define Bintohex(arg,checkendian)	(sizeof(arg)>MD5_DIGLEN?\
  (fail(#arg" too large"),terminate(EX_SOFTWARE),""):\
- bintohex(&(arg),sizeof(arg)))
+ bintohex(&(arg),sizeof(arg),checkendian))
 
 static const char*const nullp,From_[]=FROM,XFrom_[]=XFROM,cucipopn[]="cucipop",
  Received[]=
-  "Received: by %s for %s\r\n (with Cubic Circle's %s (%s) %.24s)\r\n",
- version[]=VERSION;
+  "Received: by %s (mbox %s)\r\n (with Cubic Circle's %s (%s) %.24s)\r\n",
+ version[]=VERSION,bulletins_path[]=BULLETINS_PATH;
 static char linebuf[LINEBUFLEN+1],*arg1,*arg2,user[ARGLEN+1],*host,*timestamp;
 static FILE*fmbox,*sockout;
 static size_t hostnduser;
@@ -66,20 +69,31 @@ long spoofcheck;
 static struct sockaddr_in peername;
 
 struct msg
- { off_t start;
+ { size_t order;
+   off_t start;
+   int file;
+   time_t mtime;
 #ifdef UIDL
    uidl_t uidl;
+   unsigned statskip;
 #endif
-   long virtsize;
+   long virtsize,realsize;
    unsigned char deleted,tooold;
  }*msgs;
 
-static size_t msgs_filled,msgs_max,totmsgs,transfmsgs;
+#define MSGS(i) (msgs[msgs[i].order])
+
+static size_t msgs_filled,priv_msgs_filled,msgs_max,totmsgs,transfmsgs;
 static long totsize,transfsize;
 static off_t oldend;
 static unsigned audit,berkeleyformat;
-static time_t tsabotage;
+static time_t tsabotage,tsessionstart;
 static const char*mailbox;
+static struct stateinfo si;
+
+#define TBUL_READ(x)	(si.readbulletin[(x)/8]&1<<((x)&7))
+#define SBUL_READ(x)	(si.readbulletin[(x)/8]|=1<<((x)&7))
+#define SBUL_UNREAD(x)	(si.readbulletin[(x)/8]&=~(1<<((x)&7)))
 
 static void bailout(void);
 
@@ -89,24 +103,30 @@ static void zombiecollect(void)
 
 void blocksignals(void)
 { signal(SIGHUP,SIG_IGN);signal(SIGINT,SIG_IGN);signal(SIGTERM,SIG_IGN);
-  signal(SIGPIPE,SIG_IGN);signal(SIGURG,SIG_IGN);signal(SIGALRM,SIG_IGN);
+  signal(SIGPIPE,SIG_IGN);signal(SIGALRM,SIG_IGN);
+#ifdef SIGURG
+  signal(SIGURG,SIG_IGN);
+#endif
 }
 
 void setsignals(void)
 { Signal(SIGHUP,bailout);Signal(SIGINT,bailout);Signal(SIGTERM,bailout);
-  Signal(SIGPIPE,bailout);Signal(SIGURG,bailout);Signal(SIGALRM,bailout);
+  Signal(SIGPIPE,bailout);Signal(SIGALRM,bailout);
+#ifdef SIGURG
+  Signal(SIGURG,bailout);
+#endif
 }
 
 static size_t findmsg(void)
 { size_t i;
   return arg1&&(i=strtol(arg1,(char**)0,10)-1)<msgs_filled&&
-   !(msgs[i].deleted&1)?i+1:0;
+   !(MSGS(i).deleted&1)?i+1:0;
 }
 
-static const char*bintohex(const void*ptr,size_t size)
+static const char*bintohex(const void*ptr,size_t size,const int checkendian)
 { static char hex[BIN2HEXLEN+1],*hp;int inc=1;
   { const int test=1;
-    if(!*(const char*)&test)
+    if(checkendian&&!*(const char*)&test)
        ptr=(const char*)ptr+size+(inc=-1);
   }
   for(hp=hex;size--;ptr=(const char*)ptr+inc,hp+=2)
@@ -124,12 +144,27 @@ static const char*peer(void)
 { return iptoascii(&peername);
 }
 
+void logfinal(int level,int lost,unsigned autodel)
+{ if(!*user||transfsize||totmsgs||audit)
+     syslog(level,"%s%s %s %ld, %s%ld (%ld), %ld (%ld)",lost?"lost ":"",Vuser,
+      peer(),(long)(time((time_t*)0)-tsessionstart),autodel?"-":"",
+      (long)transfmsgs,transfsize,(long)totmsgs,totsize);
+}
+
 static void terminate(int code)
 { unlockmbox();
-  if(!*user||transfsize||totmsgs||audit)
-     syslog(LOG_NOTICE,"lost %s %s, %ld (%ld), %ld (%ld)",user,peer(),
-      (long)transfmsgs,transfsize,(long)totmsgs,totsize);
-  exit(code);
+  if(transfmsgs)
+   { struct stateinfo*sip;
+     if(sip=getstate(Vuser,tsessionstart))
+	si=*sip;
+     si.lastabort=tsessionstart;
+     if(!si.lastsuccess)
+	si.lastsuccess=tsessionstart;
+     if(si.brokenstate+1>si.brokenstate)
+	si.brokenstate++;
+     putstate(Vuser,&si);
+   }
+  closedb();exitappdb();logfinal(LOG_NOTICE,1,0);exit(code);
 }
 
 static void bailout(void)
@@ -145,7 +180,7 @@ static int cmpbuf(const char*p)
 { return!strcmp(p,linebuf);
 }
 
-static int readcommand(int autodel)
+static int readcommand(unsigned autodel)
 { fflush(sockout);alarm(TIMEOUT);
   if(!fgets(linebuf,LINEBUFLEN+1,sockin))
    { if(autodel)
@@ -201,14 +236,23 @@ void wipestring(char*string)
      *string++='\0';
 }
 
+static const auth_identity*transmogrify(const auth_identity*pass)
+{ initappdb();opendb();
+  return setuid(auth_whatuid(pass))?(auth_identity*)0:pass;
+}
+
 static const auth_identity*checkpw(const char*pw)
-{ const auth_identity*pass;
+{ const auth_identity*pass;int valid=0;
   if(!(pass=auth_finduser(user,fileno(sockin)))||auth_whatuid(pass)<LOWEST_UID)
      goto ret0;
-  if(!auth_checkpassword(pass,pw,0))
+  ;{ char*secret;
+     if((secret=(char*)auth_getsecret(pass))&&*secret)
+	valid=!strcmp(pw,secret),wipestring(secret);
+   }
+  if(!valid&&!auth_checkpassword(pass,pw,0))
 ret0:
      return 0;
-  return setuid(auth_whatuid(pass))?(auth_identity*)0:pass;
+  return transmogrify(pass);
 }
 
 static const auth_identity*checkchallenge(const char*ch)
@@ -228,8 +272,8 @@ static const auth_identity*checkchallenge(const char*ch)
 	MD5Init(&context);
 	MD5Update(&context,(unsigned char*)timestamp,stlen+seclen);
 	wipestring(timestamp+stlen);MD5Final(digest,&context);
-	if(!strcmp(Bintohex(digest),ch))
-	   return setuid(auth_whatuid(pass))?(auth_identity*)0:pass;
+	if(!strcmp(Bintohex(digest,0),ch))
+	   return transmogrify(pass);
       }
    }
 ret0:
@@ -244,33 +288,44 @@ static void corrupted(off_t offset)
 
 static void addblock(const off_t start)
 { if(msgs_filled==msgs_max&&
-   !(msgs=realloc(msgs,(msgs_max+=GROWSTEP)*sizeof(struct msg))))
+   !(msgs=realloc(msgs,(msgs_max+=GROWSTEP)*sizeof*msgs)))
      outofmem();
-  msgs[msgs_filled].start=start;msgs[msgs_filled].virtsize=0;
-  msgs[msgs_filled++].deleted=0;
+  msgs[msgs_filled].order=msgs_filled;msgs[msgs_filled].start=start;
+  msgs[msgs_filled].virtsize=0;msgs[msgs_filled++].deleted=0;
 }
 
-static int openmailbox(const auth_identity*pass,const time_t agecutoff)
-{ off_t start,end;size_t virtsize,extra;uidl_t uidl;unsigned inheader;
-  long clen;int fdsockin;
-  alarm(0);clen=end=start=0;fdsockin=fileno(sockin);
-  if(!(fmbox=fopen(mailbox=auth_mailboxname((auth_identity*)pass),"r+b")))
-   { if(errno==ENOENT)
-	goto nomailbox;					 /* no mailbox found */
-     return 0;
-   }
-  ungetc(fgetc(fmbox),fmbox);				      /* prime atime */
-  ;{ struct stat stbuf;int fd;
-     fstat(fd=fileno(fmbox),&stbuf);
-     if(stbuf.st_atime-stbuf.st_mtime<32)	/* modification in progress? */
-      { if(!lockmbox(fd,mailbox))
-	   return 0;
-	fstat(fd,&stbuf);unlockmbox();setsignals();
-      }
-     oldend=end=stbuf.st_size;
-   }
+char*bullname(int bullno)
+{ static char buf[4];
+#if MAXBULLETINS>1000
+#error Increase the buf array length if you want to do this
+#endif
+  sprintf(buf,"%02d",bullno);
+  return buf;
+}
+
+FILE*openbulletin(const int extrafile,FILE*fp)
+{ if(extrafile>=0)
+     fp=fopen(bullname(extrafile),"rb");
+  return fp;
+}
+
+static void scanfile(const int extrafile,FILE*fp,const time_t agecutoff)
+{ size_t virtsize,extra,omf;int fdsockin;off_t start,end;uidl_t uidl;long clen;
+  unsigned inheader,statskip;
+  clen=end=start=statskip=0;fdsockin=-1;omf=msgs_filled;
   virtsize=extra=STRLEN(XFrom_)-STRLEN(From_)+STRLEN(Received)+hostnduser-2*2+
    STRLEN(cucipopn)-2+STRLEN(version)-2+24-5;
+  if(!(fp=openbulletin(extrafile,fp)))
+     if(extrafile<0)
+	goto nomailbox;
+     else
+	return;
+  if(extrafile<0)
+     end=oldend,fdsockin=fileno(sockin);
+  else
+   { struct stat stbuf;
+     fstat(fileno(fp),&stbuf);end=stbuf.st_size;
+   }
   fcntl(fdsockin,F_SETFL,(long)O_NONBLOCK);	/* non-blocking spoof enable */
   for(;;)
    { unsigned newlines=1;
@@ -280,22 +335,24 @@ static int openmailbox(const auth_identity*pass,const time_t agecutoff)
 	 { if(!end)
 	      goto nomailbox;
 	   end--;
-	   if(*p!=(c=getc(fmbox)))
-	    { ungetc(c,fmbox);end++;
+	   if(*p!=(c=getc(fp)))
+	    { ungetc(c,fp);end++;
 	      break;
 	    }
 	   newlines=2;
 	   if(!*++p)
-	    { inheader=1;addblock(start);msgs[msgs_filled-1].deleted=0;
-	      msgs[msgs_filled-1].tooold=1;
-	      if(msgs_filled>1)
-	       { totsize+=msgs[msgs_filled-2].virtsize=
-		  virtsize+start-msgs[msgs_filled-2].start;
+	    { struct msg*msp;
+	      inheader=1;addblock(start);(msp=msgs+msgs_filled-1)->deleted=0;
+	      msp->tooold=1;msp->file=extrafile;msp->mtime=agecutoff;
+	      if(msgs_filled-omf>1)
+	       { msp--;
+		 totsize+=msp->virtsize=virtsize+
+		  (msp->realsize=start-msp->start);
 #ifdef UIDL
-		 msgs[msgs_filled-2].uidl=uidl;
+		 msp->uidl=uidl;msp->statskip=statskip;
 #endif
 	       }
-	      uidl=tsabotage;virtsize=extra;
+	      statskip=0;uidl=tsabotage;virtsize=extra;
 #define DATELEN 20
 	      if(end>DATELEN+4)
 	       { char circ[DATELEN],circo[DATELEN+1];unsigned rot,frot;
@@ -303,17 +360,21 @@ static int openmailbox(const auth_identity*pass,const time_t agecutoff)
 		  { if(!end)
 		       goto nomailbox;
 		    end--;virtsize++;start++;
-		    if((c=getc(fmbox))=='\n')
+		    if((c=getc(fp))=='\n')
 		       break;
 		    circ[rot]=c;frot++;
 		  }
 		 if(frot>=DATELEN)
 		  { for(frot=0;frot<DATELEN;frot++,rot=(rot+1)%DATELEN)
 		       circo[frot]=circ[rot];
-		    if(atotime(circo)-agecutoff>=0)
-		     { if(tsabotage)
-			  uidl=0;
-		       msgs[msgs_filled-1].tooold=0;
+		    if(extrafile<0)
+		     { time_t mtime;struct msg*msp=msgs+msgs_filled-1;
+		       if((mtime=atotime(circo))-agecutoff>=0)
+			{ if(tsabotage)
+			     uidl=0;
+			  msp->tooold=0;
+			}
+		       msp->mtime=mtime;
 		     }
 		  }
 	       }
@@ -330,7 +391,7 @@ nomailbox:
 	      goto thatsall;
 	   if(!inheader)
 	      clen--;
-	   c=getc(fmbox);
+	   c=getc(fp);
 	   if(spoofcheck>=0&&!(++spoofcheck&SPOOFMASK))
 	    { char c;
 	      spoofcheck=0;
@@ -346,48 +407,72 @@ trynl:
 	   uidl=uidl*67067L+(unsigned char)c;
 #endif
 	   if(c=='\n')
-	    { virtsize++;
-	      if(--newlines&&inheader&&berkeleyformat)
-	       { static const char contlength[]=CONTLENGTH;const char*p;
-		 for(p=contlength;;)
+isnl:	    { virtsize++;
+	      if(--newlines&&inheader)
+	       { static const char contlength[]=CONTLENGTH,
+		  statfield[]=STATUSFIELD,xstatfield[]=XSTATUSFIELD;
+		 const char*p,*ps,*psx;unsigned skiplen;uidl_t luidl;
+		 p=berkeleyformat?0:contlength;ps=statfield;psx=xstatfield;
+		 for(skiplen=0,luidl=uidl;;)
 		  { if(!end)
 		       goto thatsall;
-		    end--;
-		    if(*p!=tolower(c=getc(fmbox)))
+		    end--;c=getc(fp);
+		    if(p&&*p!=tolower(c))
+		       p=0;
+		    if(ps&&*ps!=tolower(c))
+		       ps=0;
+		    if(psx&&*psx!=tolower(c))
+		       psx=0;
+		    if(!p&&!ps&&!psx)
+		     { uidl=luidl;
 		       goto trynl;
-		    newlines=2;
-		    if(!*++p)
+		     }
+#ifdef UIDL
+		    luidl=luidl*67067L+(unsigned char)c;
+#endif
+		    newlines=2;skiplen++;
+		    if(ps&&!*++ps||psx&&!*++psx)
+		     { statskip+=skiplen+1;
+		       do
+			{ if(!end)
+			     goto thatsall;
+			  end--;statskip++;
+			}
+		       while(fgetc(fp)!='\n');
+		       goto isnl;
+		     }
+		    if(p&&!*++p)
 		     { for(;;)
 			{ if(!end)
 			     goto thatsall;
 			  end--;
-			  switch(c=fgetc(fmbox))
+			  switch(c=fgetc(fp))
 			   { case ' ':case '\t':continue;
 			   }
 			  break;
 			}
-		       for(clen=0;;end--,c=fgetc(fmbox))
+		       for(clen=0;;end--,c=fgetc(fp))
 			{ if(c<'0'||c>'9')
 			     goto trynl;
 			  clen=((unsigned long)clen*10)+c-'0';
 			  if(!end)   /* memory requirements to a perfect fit */
-thatsall:		   { fcntl(fdsockin,F_SETFL,0L);      /* block again */
+thatsall:		   { struct msg*msp;
+			     fcntl(fdsockin,F_SETFL,0L);      /* block again */
 			     if(clen>0)
 				syslog(LOG_WARNING,"%s mailbox %ld short",
 				 mailbox,(long)clen);
-			     if(!(msgs=realloc(msgs,(msgs_max=
-			      (totmsgs=msgs_filled)+2)*sizeof(struct msg))))
-				outofmem();
-			     addblock(start+=oend);
-			     msgs[--msgs_filled].deleted=1;
-			     if(msgs_filled>0)
-			      { totsize+=msgs[msgs_filled-1].virtsize=
-				 virtsize+start-msgs[msgs_filled-1].start;
+			     msp=msgs+msgs_filled;start+=oend;msp->deleted=1;
+			     if(msgs_filled>omf)
+			      { msp--;
+				totsize+=msp->virtsize=virtsize+
+				 (msp->realsize=start-msp->start);
 #ifdef UIDL
-				msgs[msgs_filled-1].uidl=uidl;
+				msp->uidl=uidl;msp->statskip=statskip;
 #endif
 			      }
-			     return 1;
+			     if(extrafile>=0)
+				fclose(fp);
+			     return;
 			   }
 			}
 		     }
@@ -405,6 +490,97 @@ thatsall:		   { fcntl(fdsockin,F_SETFL,0L);      /* block again */
    }
 }
 
+static int mcmp(const struct msg*a,const struct msg*b)
+{ a=msgs+a->order;b=msgs+b->order;
+  switch(si.brokenstate)
+   { case 0:
+	return -(a->order<b->order);
+     case 1:
+	return -(a->realsize<b->realsize);
+     default:
+	if(tsessionstart-a->mtime<DIRECTFETCHAGE||
+	 tsessionstart-b->mtime<DIRECTFETCHAGE)
+	   return -(a->mtime>b->mtime);
+     case 2:
+      { unsigned sa=a->realsize/BSIZECATEGORY,sb=b->realsize/BSIZECATEGORY;
+	return sa<sb?-1:sa>sb?1:-(a->mtime>b->mtime);
+      }
+   }
+}
+
+static void mswap(struct msg*a,struct msg*b)
+{ int t;
+  t=a->order;a->order=b->order;b->order=t;
+}
+
+static int openmailbox(const auth_identity*pass,
+ const time_t agecutoff)
+{ FILE*fp;
+  alarm(0);
+  if(!(fmbox=fp=fopen(mailbox=auth_mailboxname((auth_identity*)pass),"r+b")))
+   { if(errno==ENOENT)
+	goto nomailbox;					 /* no mailbox found */
+     return 0;
+   }
+  ungetc(fgetc(fp),fp);					      /* prime atime */
+  ;{ struct stat stbuf;int fd;
+     fstat(fd=fileno(fp),&stbuf);
+     if(stbuf.st_atime-stbuf.st_mtime<32)	/* modification in progress? */
+      { if(!lockmbox(fd,mailbox))
+	   return 0;
+	fstat(fd,&stbuf);unlockmbox();setsignals();
+      }
+     oldend=stbuf.st_size;
+   }
+nomailbox:
+  ;{ struct stateinfo*sip;
+     if(sip=getstate(Vuser,tsessionstart))
+      { unsigned i;
+	si=*sip;
+	for(i=priv_msgs_filled;i--;)
+	 { unsigned b=msgs[i].file;
+	   if(TBUL_READ(b))
+	      if(msgs[i].mtime-si.lastsuccess>=0)
+		 SBUL_UNREAD(b);
+	      else
+	       { msgs[i]=msgs[--priv_msgs_filled];msgs[i].order=i;
+		 --msgs_filled;
+	       }
+	 }
+      }
+   }
+  scanfile(-1,fp,agecutoff);
+  if(si.brokenstate>0)
+     hsort(msgs,msgs_filled,sizeof*msgs,mcmp,mswap);
+  totmsgs=msgs_filled;
+  return 1;
+}
+
+void checkbulletins()
+{ if(!chdir(bulletins_path))
+   { struct stat stbuf;static time_t lastmods;static int nextextrafile=-1;
+     if(nextextrafile>=0)
+      { if(!stat(bullname(nextextrafile),&stbuf)&&stbuf.st_mtime>lastmods)
+	   goto scanem;
+      }
+     else
+      { int bul;
+scanem: for(msgs_filled=0,bul=MAXBULLETINS;bul--;)
+	   if(!stat(bullname(bul),&stbuf))
+	    { time_t mtime=stbuf.st_mtime;
+	      if(!lastmods||mtime-lastmods>0)
+		 lastmods=mtime;
+	      scanfile(bul,0,mtime);
+	    }
+	priv_msgs_filled=msgs_filled;
+      }
+     if(nextextrafile<=0)
+	nextextrafile=MAXBULLETINS;
+     nextextrafile--;
+   }
+  chdir(ROOTDIR);
+}
+
 static int rread(const int fd,char*a,const unsigned len)
 { int i;size_t left=len;
   do
@@ -423,11 +599,11 @@ static int rwrite(const int fd,char*a,const unsigned len)
 
 #ifdef UIDL
 static void printuidl(size_t i)
-{ size_t sz=msgs[i].virtsize-hostnduser;
-  (sizeof msgs[i].uidl+sizeof msgs[i].virtsize)*2>UIDLLEN?
+{ size_t sz=MSGS(i).virtsize-hostnduser-MSGS(i).statskip;
+  (sizeof MSGS(i).uidl+sizeof MSGS(i).virtsize)*2>UIDLLEN?
    fail("uidl too long"),terminate(EX_SOFTWARE),0:0;
-  fprintf(sockout,"%ld %s",(long)(i+1),Bintohex(msgs[i].uidl));
-  fprintf(sockout,"%s\r\n",Bintohex(sz));
+  fprintf(sockout,"%ld %s",(long)(i+1),Bintohex(MSGS(i).uidl,1));
+  fprintf(sockout,"%s\r\n",Bintohex(sz,1));
 }
 #endif
 
@@ -439,9 +615,8 @@ main(int argc,const char*const argv[])
 { unsigned loginattempts=0,quiet=0,debug=0;short port=POP3_PORT;
   unsigned douser=1,douidl=1,doapop=1,dotop=1,autodel=0,sabotage=0;
   static const char tdotnl[]=".\r\n";time_t agecutoff=-AGETOLERANCE;
-  chdir(ROOTDIR);openlog(cucipopn,LOG_PID,LOG_FACILITY);sockin=stdin;
-  sockout=stdout;
-  if(!(msgs=malloc(GROWSTEP*sizeof(struct msg))))
+  openlog(cucipopn,LOG_PID,LOG_FACILITY);sockin=stdin;sockout=stdout;
+  if(!(msgs=malloc(GROWSTEP*sizeof*msgs)))
      outofmem();
   msgs_max=GROWSTEP;
 #ifdef SIGXCPU
@@ -507,8 +682,18 @@ main(int argc,const char*const argv[])
 #ifdef UIDL
 		  " UIDL"
 #endif
-		  "\nSystem mailbox:\t\t%s\n",cucipopn,version,
-		  auth_mailboxname(auth_finduid(getuid(),0)));
+		  "\nSystem mailbox:\t\t%s\n"
+#ifdef USE_DB
+		  "Virtual pop database:\t%s\n"
+		  "Virtual pop user:\t%s\n"
+		  "Housekeeping tree:\t%s\n"
+		  "Bulletin subdirectory:\t%s\t00-%02d\n"
+		  "State database:\t\t%s/%s\n"
+#endif
+		  ,cucipopn,version,
+		  auth_mailboxname(auth_finduid(getuid(),0)),virtualhostdb,
+		  virtualuser,cplib,bulletins_path,MAXBULLETINS-1,cplib,
+		  statedbfile);
 		 return EXIT_SUCCESS;
 usg:	      default:printusg();
 		 return EX_USAGE;
@@ -530,7 +715,7 @@ usg:	      default:printusg();
       !debug&&(errno==ENOTSOCK||errno==EINVAL))
       { int serverfd,curfd;
 	signal(SIGHUP,SIG_IGN);signal(SIGPIPE,SIG_IGN);fclose(stdin);
-	fclose(stdout);serverfd=socket(AF_INET,SOCK_STREAM,TCP_PROT);
+	fclose(stdout);serverfd=socket(PF_INET,SOCK_STREAM,TCP_PROT);
 	peername.sin_family=AF_INET;peername.sin_addr.s_addr=INADDR_ANY;
 	peername.sin_port=htons(port);curfd=-1;
 	setsockopt(serverfd,SOL_SOCKET,SO_REUSEADDR,&curfd,sizeof curfd);
@@ -550,11 +735,12 @@ usg:	      default:printusg();
 	   else
 	      syslog(LOG_ERR,"can't write %s",linebuf);
 	 }
+	initappdb();exitappdb();		    /* prime database system */
 	do
 	 { close(curfd);
 	   if(namelen<0)
 	      sleep(LOCKSLEEP);
-	   namelen=sizeof peername;zombiecollect();
+	   namelen=sizeof peername;checkbulletins();zombiecollect();
 	   if((curfd=accept(serverfd,(struct sockaddr*)&peername,&namelen))<0)
 	      continue;
 	   zombiecollect();
@@ -568,9 +754,9 @@ usg:	      default:printusg();
 	      terminate(EX_PROTOCOL);
       }
      else
-	fclose(stderr);
+	checkbulletins(),fclose(stderr);
    }
-  setsignals();
+  chdir(bulletins_path);tsessionstart=time((time_t*)0);setsignals();
   ;{ struct utsname name;
      uname(&name);
      ;{ size_t hlen;
@@ -579,15 +765,13 @@ usg:	      default:printusg();
 	   outofmem();
       }
      ;{ pid_t pid=getpid();
-	sprintf(timestamp,"<%s",Bintohex(pid));
-      }
-     ;{ time_t t=time((time_t*)0);
-	if(sabotage)
-	   tsabotage=t;
-	agecutoff=t-agecutoff;
-	sprintf(timestamp+1+sizeof(pid_t)*2,"%s@%s>",Bintohex(t),
-	 strcpy(host,name.nodename));
+	sprintf(timestamp,"<%s",Bintohex(pid,1));
       }
+     if(sabotage)
+	tsabotage=tsessionstart;
+     agecutoff=tsessionstart-agecutoff;
+     sprintf(timestamp+1+sizeof(pid_t)*2,"%s@%s>",Bintohex(tsessionstart,1),
+      strcpy(host,name.nodename));
    }
   fprintf(sockout,"+OK Cubic Circle's %s POP3 ready %s"
 #ifdef LOG_FAILED_PASSWORD
@@ -635,17 +819,17 @@ usg:	      default:printusg();
 	    { strcpy(user,arg1);
 	      if(pass=checkchallenge(arg2))
 okpass:	       { free(timestamp);
-		 hostnduser=strlen(host)+strlen(user);
+		 hostnduser=strlen(host)+strlen(Vuser);
 		 if(openmailbox(pass,agecutoff))
 		    fprintf(sockout,"+OK Congratulations!\r\n");
 		 else
-		  { syslog(LOG_ALERT,"Error opening %s's mailbox",user);
+		  { syslog(LOG_ALERT,"Error opening %s's mailbox",Vuser);
 		    fail("Error opening your mailbox");terminate(EX_NOINPUT);
 		  }
 		 break;
 	       }
 	      else
-failpass:      { syslog(LOG_ERR,"authentication failure %s %s",user,peer());
+failpass:      { syslog(LOG_ERR,"authentication failure %s %s",Vuser,peer());
 		 sleep(++loginattempts);
 		 fprintf(sockout,
 		  "-ERR %s: Invalid password or username (check case)\r\n",
@@ -685,7 +869,8 @@ aluser:	   fail("Already selected a username");
       { if(arg1)
 	 { size_t i;
 	   if(i=findmsg())
-	      fprintf(sockout,"+OK %ld %ld\r\n",(long)i,msgs[i-1].virtsize);
+	      fprintf(sockout,"+OK %ld %ld\r\n",
+	       (long)i,MSGS(i-1).virtsize);
 	   else
 	      goto nomsg;
 	 }
@@ -694,8 +879,9 @@ aluser:	   fail("Already selected a username");
 	   fprintf(sockout,"+OK %ld messages (%ld octets)\r\n",(long)totmsgs,
 	    (long)totsize);
 	   for(i=0;i<msgs_filled;i++)
-	      if(!(msgs[i].deleted&1))
-		 fprintf(sockout,"%ld %ld\r\n",(long)(i+1),msgs[i].virtsize);
+	      if(!(MSGS(i).deleted&1))
+		 fprintf(sockout,"%ld %ld\r\n",(long)(i+1),
+		  MSGS(i).virtsize);
 	   goto tdot;
 	 }
       }
@@ -704,13 +890,13 @@ aluser:	   fail("Already selected a username");
       ||dotop&&arg2&&cmpbuf("top")
 #endif
       )
-      { size_t i;
-	if(i=findmsg())
+      { size_t i;FILE*fp;struct msg*pvm;
+	if((i=findmsg())&&(fp=openbulletin((pvm=&MSGS(--i))->file,fmbox)))
 	 { int newline=1,lines=-1,inbody=0;size_t dodel=0;
-	   alarm(0);fseek(fmbox,msgs[--i].start,SEEK_SET);
-	   if(STRLEN(From_)!=fread(linebuf,1,STRLEN(From_),fmbox)||
+	   alarm(0);fseek(fp,pvm->start,SEEK_SET);
+	   if(STRLEN(From_)!=fread(linebuf,1,STRLEN(From_),fp)||
 	    memcmp(From_,linebuf,STRLEN(From_)))
-	      corrupted(msgs[i].start);
+	      corrupted(pvm->start);
 	   if(arg2)
 	      lines=strtol(arg2,(char**)0,10);
 	   else
@@ -718,19 +904,18 @@ aluser:	   fail("Already selected a username");
 	      if(autodel)
 		 dodel=i+1;
 	    }
-	   fprintf(sockout,"+OK %ld octets%s\r\n",msgs[i].virtsize,
+	   fprintf(sockout,"+OK %ld octets%s\r\n",pvm->virtsize,
 	    dodel?" (autodeleted)":"");
-	   transfsize+=msgs[i].virtsize;
-	   i=msgs[i+1].start-msgs[i].start-STRLEN(From_);
+	   transfsize+=pvm->virtsize;i=pvm->realsize-STRLEN(From_);
 	   ;{ time_t lasttime;
 	      lasttime=time((time_t*)0);
-	      fprintf(sockout,Received,host,user,cucipopn,version,
+	      fprintf(sockout,Received,host,Vuser,cucipopn,version,
 	       ctime(&lasttime));
 	    }
 	   fprintf(sockout,XFrom_);
 	   while(i--)
 	    { unsigned c;
-	      switch(c=getc(fmbox))
+	      switch(c=getc(fp))
 	       { case '\n':
 		    if(newline)
 		       inbody=1;
@@ -738,7 +923,7 @@ aluser:	   fail("Already selected a username");
 		     { if(autodel)
 prequitstate:		{ autodel=2;
 			  for(i=0;i<msgs_filled;i++)
-			     msgs[i].deleted&=2;
+			     MSGS(i).deleted&=2;
 			  goto quitstate;
 			}
 		       terminate(EX_PROTOCOL);
@@ -757,8 +942,11 @@ prequitstate:		{ autodel=2;
 	      putc(c,sockout);
 	    }
 	   fprintf(sockout,tdotnl);
-	   if(dodel&&!fflush(sockout)&&msgs[dodel-1].tooold)
-	      msgs[dodel-1].deleted|=2;
+	   if(dodel&&MSGS(dodel-1).file<0&&MSGS(dodel-1).tooold&&
+	    !fflush(sockout))
+	      MSGS(dodel-1).deleted|=2;
+	   if(fp!=fmbox)
+	      fclose(fp);
 	 }
 	else
 	   goto nomsg;
@@ -767,7 +955,7 @@ prequitstate:		{ autodel=2;
       { size_t i;
 	if(i=findmsg())
 	 { fprintf(sockout,"+OK Message %ld deleted\r\n",(long)i--);
-	   msgs[i].deleted|=1;totmsgs--;totsize-=msgs[i].virtsize;
+	   MSGS(i).deleted|=1;totmsgs--;totsize-=MSGS(i).virtsize;
 	 }
 	else
 	   goto nomsg;
@@ -775,7 +963,7 @@ prequitstate:		{ autodel=2;
 #ifdef LAST_HACK
      else if(!arg1&&cmpbuf("last"))	      /* LAST does not exist anymore */
       { fprintf(sockout,"+OK 0\r\n");
-	syslog(LOG_DEBUG,"Faked LAST for %s",user);
+	syslog(LOG_DEBUG,"Faked LAST for %s",Vuser);
       }
 #endif
      else if(!arg1&&cmpbuf("noop"))
@@ -783,8 +971,8 @@ prequitstate:		{ autodel=2;
      else if(!arg1&&cmpbuf("rset"))
       { size_t i;
 	for(i=0;i<msgs_filled;i++)
-	   if(msgs[i].deleted&1)
-	      totmsgs++,totsize+=msgs[i].virtsize,msgs[i].deleted&=2;
+	   if(MSGS(i).deleted&1)
+	      totmsgs++,totsize+=MSGS(i).virtsize,MSGS(i).deleted&=2;
 	fprintf(sockout,"+OK %ld messages (%ld octets)\r\n",(long)totmsgs,
 	 (long)totsize);
       }
@@ -792,7 +980,7 @@ prequitstate:		{ autodel=2;
       { if(arg1)
 	 { size_t i;
 	   if(i=findmsg())
-#ifdef UIDL				 /* UIDL will change with endianness */
+#ifdef UIDL
 	      fprintf(sockout,"+OK "),printuidl(i-1);
 #else
 notimpl:      fail("Command not implemented");
@@ -808,7 +996,7 @@ nomsg:	      fail("No such message");
 	   fprintf(sockout,
 	    "+OK But remember to DELETE messages REGULARLY\r\n");
 	   for(i=0;i<msgs_filled;i++)
-	      if(!(msgs[i].deleted&1))
+	      if(!(MSGS(i).deleted&1))
 		 printuidl(i);
 #endif
 tdot:	   fprintf(sockout,tdotnl);
@@ -827,17 +1015,19 @@ tdot:	   fprintf(sockout,tdotnl);
    }
 quitstate:
   ;{ off_t target=0;size_t i;
-     for(i=0;i<msgs_filled&&!msgs[i].deleted;i++);
+     for(i=priv_msgs_filled;i<msgs_filled&&!msgs[i].deleted;i++);
      if(i<msgs_filled)
       { char*buf;off_t source;size_t slen;int fd=fileno(fmbox);
 	alarm(0);setbuf(sockin,(char*)0);setbuf(fmbox,(char*)0);free(host);
 	if(!(buf=malloc(COPYBUF)))
 	   outofmem();
 	if(!lockmbox(fd,mailbox))
-	 { syslog(LOG_ALERT,"Error locking %s's mailbox",user);
+	 { syslog(LOG_ALERT,"Error locking %s's mailbox",Vuser);
 	   fail("Error locking your mailbox");terminate(EX_NOINPUT);
 	 }
-	msgs[msgs_filled++].deleted=0;
+	;{ struct msg*msp=msgs+msgs_filled-1;
+	   addblock(msp->start+msp->realsize);
+	 }
 	if(oldend>(source=lseek(fd,(off_t)0L,SEEK_END)))
 	   goto corrupt;
 	addblock(source);msgs[--msgs_filled].deleted=1;
@@ -871,17 +1061,28 @@ jin:		    source+=readin;lseek(fd,target,SEEK_SET);
 	   goto writerr;
 	unlockmbox();
 	if(close(fd))
-writerr: { syslog(LOG_CRIT,"write error mailbox %s at %ld",user,(long)target);
+writerr: { syslog(LOG_CRIT,"write error mailbox %s at %ld",Vuser,(long)target);
 	   fail("Error writing mailbox");terminate(EX_OSERR);
 	 }
       }
-     if(!quiet&&(transfsize||totmsgs||audit))
-	syslog(LOG_INFO,"%s%s %s, %s%ld (%ld), %ld (%ld)",
-	 autodel==2?"lost ":"",user,peer(),autodel?"-":"",(long)transfmsgs,
-	 transfsize,(long)totmsgs,totsize);
+     if(!quiet)
+	logfinal(LOG_INFO,autodel==2,autodel);
      fprintf(sockout,"+OK Was it as good for you, as it was for me?  (");
      fprintf(sockout,target?">%ld bytes left)\r\n":
       "clean as a baby)\r\n",(long)target);
    }
+  ;{ unsigned i;
+     for(i=priv_msgs_filled;i--;)
+	if(msgs[i].deleted&1)
+	 { unsigned b=msgs[i].file;
+	   SBUL_READ(b);
+	 }
+   }
+  si.lastsuccess=tsessionstart;
+  if(transfmsgs)
+     si.brokenstate=0;
+  if(!si.lastabort)
+     si.lastabort=tsessionstart;
+  blocksignals();putstate(Vuser,&si);closedb();exitappdb();
   return EXIT_SUCCESS;
 }
