@@ -1,7 +1,10 @@
-/*$Id: cucipop.h,v 1.2 1996/10/30 16:24:31 srb Exp $*/
+/*$Id: cucipop.h,v 1.3 1997/08/01 01:00:33 srb Exp $*/
 
 void castlower(char*string);
 void blocksignals(void);
 void setsignals(void);
 
+extern FILE*sockin;
+extern long spoofcheck;
+
 #define STRLEN(x)	(sizeof(x)-1)
