@@ -6,7 +6,7 @@
  ************************************************************************/
 #ifdef RCS
 static /*const*/char rcsid[]=
- "$Id: locking.c,v 1.6 1997/08/01 01:00:33 srb Exp $";
+ "$Id: locking.c,v 1.7 1998/05/13 16:57:40 srb Exp $";
 #endif
 #include "config.h"
 
@@ -37,6 +37,10 @@ static int fdlock(int fd);
 static int fdunlock(void);
 #endif
 
+#ifndef tell
+#define tell(fd)	lseek(fd,0L,SEEK_CUR)
+#endif
+
 static char*lastlocked;
 static const char lockext[]=".lock";
 
