@@ -100,8 +100,10 @@ static int unique(full,p,mode)const char*const full;char*const p;
    }
   while(0>(i=copen(full,O_WRONLY|O_CREAT|O_EXCL|O_SYNC,mode))&&errno==EEXIST&&
    retry);	    /* casually check if it already exists (highly unlikely) */
+  write(i,"0",1);					/* put in a fake pid */
   if(i<0)
-   { log("Error while writing to \"");log(full);log("\"\n");return 0;
+   { log("Error while writing to \"");log(full);log("\"\n");
+     return 0;
    }
   close(i);return 1;
 }
